import collections
from .base import BaseItertool
from .iter_dispatch import iter_


class repeat(BaseItertool):
    """
    repeat(object [,times]) -> create an iterator which returns the object
    for the specified number of times.  If not specified, returns the object
    endlessly.
    """
    def __init__(self, obj, times=None):
        self._obj = obj
        self._times = times
        self._times_called = 0

    def __next__(self):
        if self._times is None:
            return self._obj
        else:
            if self._times > self._times_called:
                self._times_called += 1
                return self._obj
            else:
                raise StopIteration


class chain(BaseItertool):
    """
    chain(*iterables) --> chain object

    Return a chain object whose .__next__() method returns elements from the
    first iterable until it is exhausted, then elements from the next
    iterable, until all of the iterables are exhausted.
    """
    def __init__(self, *iterables):
        self._iterables = iter_(iterables)
        self._current = repeat(None, 0)

    def __next__(self):
        try:
            return next(self._current)
        except StopIteration:
            self._current = iter_(next(self._iterables))
        return next(self)

    @classmethod
    def from_iterable(cls, iterable):
        obj = cls()
        obj._iterables = iter_(iterable)
        return obj


class compress(BaseItertool):
    """compress(data, selectors) --> iterator over selected data

    Return data elements corresponding to true selector elements.
    Forms a shorter iterator from selected data elements using the
    selectors to choose the data elements.
    """
    def __init__(self, data, selectors):
        self._data = iter_(data)
        self._selectors = iter_(selectors)

    def __next__(self):
        # We terminate on the shortest input sequence, so leave
        # StopIteration uncaught here.
        data = next(self._data)
        selector = next(self._selectors)
        while not bool(selector):
            data = next(self._data)
            selector = next(self._selectors)
        return data


class count(BaseItertool):
    """count(start=0, step=1) --> count object

    Return a count object whose .__next__() method returns consecutive values.
    """
    def __init__(self, start=0, step=1):
        self._n = start
        self._step = step

    def __next__(self):
        n = self._n
        self._n += self._step
        return n


class cycle(BaseItertool):
    """cycle(iterable) --> cycle object

    Return elements from the iterable until it is exhausted.
    Then repeat the sequence indefinitely.
    """
    def __init__(self, iterable):
        self._iterable = iter_(iterable)
        self._exhausted = False
        self._elements = collections.deque()

    def __next__(self):
        if not self._exhausted:
            try:
                value = next(self._iterable)
            except StopIteration:
                self._exhausted = True
                return next(self)
            self._elements.append(value)
        else:
            if len(self._elements) == 0:
                raise StopIteration
            value = self._elements.popleft()
            self._elements.append(value)
        return value


class accumulate(BaseItertool):
    """accumulate(iterable[, func]) --> accumulate object

    Return series of accumulated sums (or other binary function results).
    """
    def __init__(self, iterable, func=None):
        self._iter = iter_(iterable)
        self._func = func
        self._initialized = False
        self._accumulated = None

    def _combine(self, value):
        if self._func is not None:
            return self._func(self._accumulated, value)
        else:
            return self._accumulated + value

    def __next__(self):
        value = next(self._iter)
        if not self._initialized:
            self._accumulated = value
            self._initialized = True
        else:
            self._accumulated = self._combine(value)
        return self._accumulated



import os.path

from pkg_resources import get_distribution, DistributionNotFound

from .filter import ifilter, ifilterfalse, takewhile, dropwhile
from .grouping import groupby
from .iter_dispatch import (
    iter_, ordered_sequence_iterator, file_iterator, range_iterator
)
from .map_zip import imap, starmap, izip, izip_longest
from .permutations import (
    permutations, combinations, combinations_with_replacement
)
from .product import product  # noqa
from .range import xrange  # noqa
from .simple import accumulate, chain, compress, count, cycle, repeat
from .slicing import islice
from .tee import tee

# Python 3 equivalents.
filter = ifilter
filterfalse = ifilterfalse
zip = izip
zip_longest = izip_longest
# Remove after bartvm/fuel has been updated to use this version.
_iter = iter_


try:
    DIST = get_distribution('picklable_itertools')
    DIST_LOC = os.path.normcase(DIST.location)
    HERE = os.path.normcase(__file__)
    if not HERE.startswith(os.path.join(DIST_LOC, 'picklable_itertools')):
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'not installed'
else:
    __version__ = DIST.version


__all__ = ['ifilter', 'ifilterfalse', 'takewhile', 'dropwhile', 'groupby',
           '_iter', 'ordered_sequence_iterator', 'file_iterator',
           'range_iterator', 'imap', 'starmap', 'izip', 'izip_longest',
           'permutations', 'combinations', 'combinations_with_replacement',
           'accumulate', 'chain', 'compress', 'count', 'cycle', 'repeat',
           'islice', 'tee', 'filter', 'filterfalse', 'zip', 'zip_longest']



import sys

from .base import BaseItertool
from .iter_dispatch import iter_


class islice(BaseItertool):
    """islice(iterable, stop) --> islice object
    islice(iterable, start, stop[, step]) --> islice object

    Return an iterator whose next() method returns selected values from an
    iterable.  If start is specified, will skip all preceding elements;
    otherwise, start defaults to zero.  Step defaults to one.  If
    specified as another value, step determines how many values are
    skipped between successive calls.  Works like a slice() on a list
    but returns an iterator.
    """
    def __init__(self, iterable, start, stop=None, step=1):
        if stop is None:
            start, stop = 0, start
        if (not 0 <= start <= sys.maxsize or
                not 0 <= stop <= sys.maxsize or
                not 0 <= step <= sys.maxsize):
            raise ValueError("Indices for islice() must be None or an "
                             "integer: 0 <= x <= maxint.")

        self._iterable = iter_(iterable)
        i = 0
        while i < start:
            try:
                next(self._iterable)
                i += 1
            except StopIteration:
                break

        self._stop = stop - start
        self._step = step
        self._n = 0

    def __next__(self):
        while self._n % self._step and self._n < self._stop:
            next(self._iterable)
            self._n += 1
        if self._n == self._stop:
            raise StopIteration
        value = next(self._iterable)
        self._n += 1
        return value



"""Support code for implementing `tee`."""
import collections
import six
from picklable_itertools import iter_


class tee_iterator(six.Iterator):
    """An iterator that works in conjunction with a `tee_manager`."""
    def __init__(self, deque, manager):
        self._deque = deque
        self._manager = manager

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._deque) > 0:
            return self._deque.popleft()
        else:
            self._manager.advance()
            assert len(self._deque) > 0
            return next(self)


class tee_manager(object):
    """An object that manages a base iterator and publishes results to
    one or more client `tee_iterators`.
    """
    def __init__(self, iterable, n=2):
        self._iterable = iter_(iterable)
        self._deques = tuple(collections.deque() for _ in range(n))

    def iterators(self):
        return tuple(tee_iterator(deque, self) for deque in self._deques)

    def advance(self):
        """Advance the base iterator, publish to constituent iterators."""
        elem = next(self._iterable)
        for deque in self._deques:
            deque.append(elem)


def tee(iterable, n=2):
    """tee(iterable, n=2) --> tuple of n independent iterators."""
    return tee_manager(iter_(iterable), n=n).iterators()
